#!/usr/bin/perl

#
# Copyright (C) 2009 by Mateusz Hromada
#

use feature ":5.10";

use constant SYSCONFDIR => '/etc/kvmsh';
use constant RUNDIR => '/var/run/kvmsh';
use constant LOGFILE => '/var/log/kvmsh';

use constant VERSION => '20090830-1';

sub warn_if {
  my ($condition, $message) = @_;
  warn "$message" if( $condition );
}

sub fatal {
  my($message, $returnValue) = @_;
  warn "$message";
  exit $returnValue;
}

sub fatal_if {
  my ($condition, $message, $returnValue) = @_;
  fatal("$message", $returnValue) if( $condition );
}

sub format_size_human {
  my ($size) = @_;
  my @prefix = ('B', 'K', 'M', 'G', 'T', 'P');
  my $base = 1024;
  my $maxDepth = $#prefix;
  my $i = 0;
  while( $size >= $base && $i < $maxDepth) {
    $size /= $base;
    $i++;
  }
  return sprintf("%01.2f%s", $size, $prefix[$i]);
}

sub format_size {
  my ($size, $format) = @_;
  return $size if( $size == 0 );
  return format_size_human($size) if( $format eq 'h' );
  my $i = 0;
  my %prefix = map { $_ => $i++ } ( 'b', 'k', 'm', 'g', 't', 'p' );
  my $base = 1024;
  my $depth = $prefix{$format};
  for( my $i=0; $i < $depth; $i++) {
    $size -= $size % $base;
    $size /= $base;
  }
  $size = 1 if( $size == 0 );
  return $size;
}

sub list_auto_configs {
  my $autoDir = SYSCONFDIR.'/auto';
  opendir(AUTODIR, $autoDir);
  my @files = grep { /^\w+\.conf$/ && -f "$autoDir/$_" } readdir(AUTODIR);
  close(AUTODIR);
  return @files;
}

sub list_host_configs {
  my $hostsDir = SYSCONFDIR.'/hosts';
  opendir(HOSTSDIR, $hostsDir);
  my @files = grep { /^\w+\.conf$/ && -f "$hostsDir/$_" } readdir(HOSTSDIR);
  closedir(HOSTSDIR);
  return @files;
}

sub list_network_configs {
  my $networkDir = SYSCONFDIR.'/network';
  opendir(NETWORKDIR, $networkDir);
  my @files = grep { /^\w+$/ && -f "$networkDir/$_" } readdir(NETWORKDIR);
  closedir(NETWORKDIR);
  return @files;
}

sub check_host_name {
  my ($hostName) = @_;
  return 1 if( $hostName =~ /^\w+$/ );
  return 0;
}

sub check_network_name {
  my ($networkName) = @_;
  return 1 if( $hostName =~ /^\w+$/ );
  return 0;
}

sub check_auto_config {
  my ($hostName) = @_;
  return 1 if( grep { /$hostName\.conf$/ } list_auto_configs );
  return 0;
}

sub check_host_config {
  my ($hostName) = @_;
  return 1 if( grep { /$hostName\.conf$/ } list_host_configs );
  return 0;
}

sub check_host_running {
  my ($hostName) = @_;
  my $pidFile = RUNDIR."/hosts/$hostName.pid";
  if( -f $pidFile ) {
    open(PIDFILE, '<', $pidFile);
    my $pid = <PIDFILE>;
    close(PIDFILE);
    chomp $pid;
    my $ps = qx(ps -p $pid -o command | tail -1); 
    if( $ps =~ /^kvm -name $hostName\b.*$/ ) {
      return 1;
    }
  }
  return 0;
}

sub monitor_write {
  my ($hostName, $command) = @_;
  my $monitorFile = RUNDIR."/hosts/$hostName.monitor";
  system("echo $command | socat STDIO UNIX-CONNECT:$monitorFile > /dev/null 2> /dev/null");
}

sub monitor_read {
  my ($hostName, $command) = @_;
  my $monitorFile = RUNDIR."/hosts/$hostName.monitor";
  return qx(echo $command | socat STDIO UNIX-CONNECT:$monitorFile 2> /dev/null);
}

sub read_host_config
{
  my ($fileName) = @_;
  my $hostsDir = SYSCONFDIR.'/hosts';
  $fileName = "$hostsDir/$fileName";
  unless( -f $fileName ) {
    warn "No such file: $fileName\n";
    return undef;
  }
  foreach( keys %HOST:: )
  {
    local *symbol = "HOST::$_";
    undef $symbol;
    undef @symbol;
    undef %symbol;
  }
  {
    package HOST;
    my $return = do $fileName;
    unless( $return ) {
      warn "Couldn't parse $fileName: $@" if $@;
      warn "Couldn't do $fileName: $!"    unless defined $return;
      warn "Couldn't run $fileName"       unless $return;
    }
    $cpus = 1 unless( defined $cpus);
    $mem = 128 unless( defined $mem);
    $mouse = 'tablet' unless( defined $mouse && ( $mouse eq 'tablet' || $mouse eq 'ps2' ) );
    $keyboard = 'en-us' unless( defined $keyboard );
    $rtc = 'utc' unless( defined $rtc && ( $rtc eq 'utc' || $rtc eq 'localtime' ) );
  }
  return $HOST::name if( $HOST::name );
  return undef;
}

sub cmd_show_config
{
  my ($hostName) = @_;
  fatal_if(!check_host_name($hostName), "Incorrect host name: $hostName\n", 1);
  my $return = read_host_config("$hostName.conf");
  unless( defined $return ) {
    warn "No such host: $hostName\n";
    exit 1;
  }
  print "Host name        : $HOST::name\n";
  print "CPUs             : $HOST::cpus\n";
  print "Memory           : $HOST::mem\n";
  print "Mouse            : $HOST::mouse\n";
  print "Keyboard layout  : $HOST::keyboard\n";
  print "RTC              : $HOST::rtc\n";
  unless( defined @HOST::ide && @HOST::ide > 0 ) {
    print "IDE devices      : none\n";
  } else {
    my $prefix = 'IDE devices      :';
    my $dev = shift(@HOST::ide);
    do {
      my ($file, $idx, $media, $format, $boot) = ($dev->{'file'},
                                                  $dev->{'idx'},
                                                  $dev->{'media'},
                                                  $dev->{'format'},
                                                  $dev->{'boot'}
                                                 );
      print "$prefix Index=$idx Media=$media Format=$format Boot=$boot File=$file\n";
      $prefix = '                  ';
    } while( $dev = shift(@HOST::ide) );
  }
  unless( defined @HOST::scsi && @HOST::scsi > 0 ) {
    print "SCSI devices     : none\n";
  } else {
    my $prefix = 'SCSI devices     :';
    my $dev = shift(@HOST::scsi);
    do {
      my ($file, $idx, $media, $format, $boot) = ($dev->{'file'},
                                                  $dev->{'idx'},
                                                  $dev->{'media'},
                                                  $dev->{'format'},
                                                  $dev->{'boot'}
                                                 );
      print "$prefix Index=$idx Media=$media Format=$format Boot=$boot File=$file\n";
      $prefix = '                  ';
    } while( $dev = shift(@HOST::scsi) );
  }
  unless( defined @HOST::virtio && @HOST::virtio > 0 ) {
    print "Virtio devices   : none\n";
  } else {
    my $prefix = 'Virtio devices   :';
    my $dev = shift(@HOST::virtio);
    do {
      my ($file, $idx, $media, $format, $boot) = ($dev->{'file'},
                                                  $dev->{'idx'},
                                                  $dev->{'media'},
                                                  $dev->{'format'},
                                                  $dev->{'boot'}
                                                 );
      print "$prefix Index=$idx Media=$media Format=$format Boot=$boot File=$file\n";
      $prefix = '                  ';
    } while( $dev = shift(@HOST::virtio) );
  }
  unless( defined @HOST::net && @HOST::net > 0 ) {
    print "Network devices  : none\n";
  } else {
    my $prefix = 'Network devices  :';
    my $dev = shift(@HOST::net);
    do {
      my ($mac, $vlan, $type, $model) = ($dev->{'mac'},
                                         $dev->{'vlan'},
                                         $dev->{'type'},
                                         $dev->{'model'}
                                        );
      $model = 'rtl8139' unless( $model );
      print "$prefix Mac=$mac Vlan=$vlan Type=$type Model=$model\n";
      $prefix = '                  ';
    } while( $dev = shift(@HOST::net) );
  }
}

sub cmd_list_host {
  print "Hosts:\n";
  my @hosts = list_host_configs;
  my $nameLength = 4;
  my $memLength = 6;
  foreach my $host ( @hosts ) {
    read_host_config($host);
    $nameLength = length($HOST::name) if( length($HOST::name) > $nameLength );
    $memLength = length("$HOST::mem") if( length("$HOST::mem") > $memLength );
  }
  printf('  %-'.($nameLength+2).'s%-7s%'.($memLength+1).'s%6s'."\n", 'Host', 'State', 'Memory', 'CPUs');
  foreach my $host ( sort @hosts ) {
    read_host_config($host);
    printf('  %-'.($nameLength+2).'s%-7s%'.($memLength+1).'s%6s'."\n",
      $HOST::name,
      (check_host_running($HOST::name)?'r    ':'-    '),
      $HOST::mem,
      $HOST::cpus
    );
  }
}

sub cmd_list_auto {
  print "Auto:\n";
  my @autos = list_auto_configs;
  foreach my $auto ( sort @autos ) {
    $auto =~ s/\.conf$//;
    print "  $auto\n";
  }
}


sub cmd_list_networks {
  print "Networks:\n";
  my @networks = list_network_configs;
  foreach my $network ( sort @networks ) {
    print "  $network\n";
  }
}

sub cmd_list {
  my ($type) = @_;
  if( $type eq '-network' || $type eq '-host' || $type eq '-auto' ) {
    cmd_list_host if( $type eq '-host' );
    cmd_list_auto if( $type eq '-auto' );
    cmd_list_networks if( $type eq '-network' );
  } else {
    cmd_list_host;
    cmd_list_auto;
    cmd_list_networks;
  }
}

sub cmd_create {
  my ($hostName) = @_;
  fatal_if(!check_host_name($hostName), "Incorrect host name: $hostName\n", 1);
  fatal_if(!check_host_config($hostName), "No such config: $hostName\n", 1);
  fatal_if(check_host_running($hostName), "Host exists.\n", 1);
  read_host_config("$hostName.conf");

  my $networkDir  = SYSCONFDIR."/network";

  my $vncFile     = RUNDIR."/hosts/$HOST::name.vnc";
  my $serialFile  = RUNDIR."/hosts/$HOST::name.serial,server,nowait";
  my $monitorFile = RUNDIR."/hosts/$HOST::name.monitor,server,nowait";
  my $pidFile     = RUNDIR."/hosts/$HOST::name.pid";

  my $cmdline = "kvm";
  $cmdline .= " -name $HOST::name";
  $cmdline .= " -smp $HOST::cpus";
  $cmdline .= " -m $HOST::mem";
  $cmdline .= " -mem-path $HOST::hugepages" if( defined $HOST::hugepages );
  $cmdline .= " -usb -usbdevice tablet" if( $HOST::mouse eq 'tablet' );
  $cmdline .= " -k $HOST::keyboard";
  $cmdline .= " -localtime" if( $HOST::rtc eq 'localtime' );
  $cmdline .= " -std-vga";
  if( defined @HOST::ide && @HOST::ide > 0 ) {
    foreach my $dev ( @HOST::ide ) {
      my ($file, $idx, $media, $format, $boot) = ($dev->{'file'},
                                                  $dev->{'idx'},
                                                  $dev->{'media'},
                                                  $dev->{'format'},
                                                  $dev->{'boot'}
                                                 );
      $cmdline .= " -drive file=$file,if=ide,index=$idx,media=$media,format=$format,boot=$boot"
    }
  }
  if( defined @HOST::scsi && @HOST::scsi > 0 ) {
    foreach my $dev ( @HOST::scsi ) {
      my ($file, $idx, $media, $format, $boot) = ($dev->{'file'},
                                                  $dev->{'idx'},
                                                  $dev->{'media'},
                                                  $dev->{'format'},
                                                  $dev->{'boot'}
                                                 );
      $cmdline .= " -drive file=$file,if=scsi,index=$idx,media=$media,format=$format,boot=$boot"
    }
  }
  if( defined @HOST::virtio && @HOST::virtio > 0 ) {
    foreach my $dev ( @HOST::virtio ) {
      my ($file, $idx, $media, $format, $boot) = ($dev->{'file'},
                                                  $dev->{'idx'},
                                                  $dev->{'media'},
                                                  $dev->{'format'},
                                                  $dev->{'boot'}
                                                 );
      $cmdline .= " -drive file=$file,if=virtio,index=$idx,media=$media,format=$format,boot=$boot"
    }
  }
  if( defined @HOST::net && @HOST::net > 0 ) {
    foreach my $dev ( @HOST::net ) {
      my ($mac, $vlan, $type, $model) = ($dev->{'mac'}, $dev->{'vlan'}, $dev->{'type'}, $dev->{'model'});
      $model = 'rtl8139' unless( $model );
      $cmdline .= " -net 'nic,vlan=$vlan,macaddr=$mac,model=$model' -net 'tap,vlan=$vlan,script=$networkDir/$type'";
    }
  }
  $cmdline .= " -vnc 'unix:$vncFile'";
  $cmdline .= " -serial 'unix:$serialFile,server,nowait'";
  $cmdline .= " -monitor 'unix:$monitorFile,server,nowait'";
  $cmdline .= " -pidfile '$pidFile'";
  $cmdline .= " -daemonize";
  $return = system($cmdline);
}

sub cmd_destroy {
  my ($hostName) = @_;
  fatal_if(!check_host_name($hostName), "Incorrect host name: $hostName\n", 1);
  fatal_if(!check_host_running($hostName), "Host does not exists.\n", 1);
  monitor_write($hostName, 'quit');
}

sub cmd_shutdown {
  my ($hostName) = @_;
  fatal_if(!check_host_name($hostName), "Incorrect host name: $hostName\n", 1);
  fatal_if(!check_host_running($hostName), "Host does not exists.\n", 1);
  monitor_write($hostName, 'system_powerdown');
}

sub cmd_reboot {
  my ($hostName) = @_;
  fatal_if(!check_host_name($hostName), "Incorrect host name: $hostName\n", 1);
  fatal_if(!check_host_running($hostName), "Host does not exists.\n", 1);
  monitor_write($hostName, 'system_reset');
}

sub cmd_pause {
  my ($hostName) = @_;
  fatal_if(!check_host_name($hostName), "Incorrect host name: $hostName\n", 1);
  fatal_if(!check_host_running($hostName), "Host does not exists.\n", 1);
  monitor_write($hostName, 'stop');
}

sub cmd_unpause {
  my ($hostName) = @_;
  fatal_if(!check_host_name($hostName), "Incorrect host name: $hostName\n", 1);
  fatal_if(!check_host_running($hostName), "Host does not exists.\n", 1);
  monitor_write($hostName, 'cont');
}

sub cmd_edit {
  my ($hostName) = @_;
  my $hostsDir = SYSCONFDIR.'/hosts';
  fatal_if(!check_host_name($hostName), "Incorrect host name: $hostName\n", 1);
  my $editor = 'editor';
  $editor = $ENV{'EDITOR'} if( $ENV{'EDITOR'} );
  system("$editor $hostsDir/$hostName.conf");
}

sub dev_block_list {
  my ($hostName) = @_;
  read_host_config("$hostName.conf");
  my @devices = ();
  my @list = split(/\n/, monitor_read($hostName, "info block"));
  my @stats = split(/\n/, monitor_read($hostName, "info blockstats"));
  @list = @list[2 .. $#list-1];
  @stats = @stats[2 .. $#stats-1];
  foreach my $listLine (@list) {
    my ($devName) = ($listLine =~ /^(.*):/);
    my ($media) = ($listLine =~ / type=([^ ]*)\s/);
    my ($format) = ($listLine =~ / drv=([^ ]*)\s/);
    my ($file) = ($listLine =~ / file=([^ ]*)\s/);
    my ($removable) = ($listLine =~ / removable=(.)\s/);
    my ($locked) = ($listLine =~ / locked=(.)\s/);
    my ($inserted) = ($listLine =~ /\[not inserted\]/);
    my ($interface) = ($devName =~ /^(ide|scsi|virtio)/);
    if( $media eq 'hd' ) {
      $media = 'disk';
    }
    unless( $media eq 'disk' or $media eq 'cdrom' )
    {
      next;
    }
    $format = 'auto' unless( defined $format );
    $inserted = 0 if( $inserted eq 1 );
    $inserted = 1 if( not defined $inserted and $removable eq 1);
    my $idx = 0;
    my $busNumber = 0;
    my $deviceNumber = 0;
    my $name = $devName;
    given( $interface ) {
      when( "ide" )     {
                          ($busNumber, $deviceNumber) = ($devName =~ /^ide(\d+)-(?:cd|hd)(\d+)/);
                          $idx = 2 * $busNumber + $deviceNumber;
                          foreach my $ideDevice ( @HOST::ide ) {
                            if( defined $ideDevice->{'name'} and $ideDevice->{'idx'} == $idx ) {
                              $name = $ideDevice->{'name'};
                            }
                          }
                        }
      when( "scsi" )    {
                          ($busNumber, $deviceNumber) = ($devName =~ /^scsi(\d+)-(?:cd|hd)(\d+)/);
                          $idx = 7 * $busNumber + $deviceNumber;
                          foreach my $scsiDevice ( @HOST::scsi ) {
                            if( defined $scsiDevice->{'name'} and $scsiDevice->{'idx'} == $idx ) {
                              $name = $scsiDevice->{'name'};
                            }
                          }
                        }
      when( "virtio" )  {
                          ($deviceNumber) = ($devName =~ /^virtio(\d*)/);
                          $idx = $deviceNumber;
                          foreach my $virtioDevice ( @HOST::virtio ) {
                            if( defined $virtioDevice->{'name'} and $virtioDevice->{'idx'} == $idx ) {
                              $name = $virtioDevice->{'name'};
                            }
                          }
                        }
    }
    my ($statsLine) = grep { /^$devName:/ } @stats;
    my ($readBytes) = ($statsLine =~ / rd_bytes=(\d+)\s/);
    my ($writeBytes) = ($statsLine =~ / wr_bytes=(\d+)\s/);
    my ($readOpers) = ($statsLine =~ / rd_operations=(\d+)\s/);
    my ($writeOpers) = ($statsLine =~ / wr_operations=(\d+)\s/);
    push(@devices, { name => $name,
                     devName => $devName,
                     media => $media,
                     format => $format,
                     file => $file,
                     removable => $removable,
                     locked => $locked,
                     inserted => $inserted,
                     interface => $interface,
                     idx => $idx,
                     busNumber => $busNumber,
                     deviceNumber => $deviceNumber,
                     readBytes => $readBytes,
                     writeBytes => $writeBytes,
                     readOpers => $readOpers,
                     writeOpers => $writeOpers
                   });
  }
  return @devices;
}

sub cmd_block_list {
  my ($hostName) = @_;
  fatal_if(!check_host_name($hostName), "Incorrect host name: $hostName\n", 1);
  fatal_if(!check_host_running($hostName), "Host does not exists.\n", 1);
  my @devices = dev_block_list($hostName);
  @devices = sort { $a->{'idx'} cmp $b->{'idx'} } @devices;
  $nameLength = 6;
  $indexLength = $devices[-1]->{'idx'};
  $indexLength = length($indexLength);
  $indexLength = 5 if( $indexLength < 5 );
  foreach my $device ( @devices ) {
    if( length($device->{'name'}) > $nameLength ) {
      $nameLength = length($device->{'name'});
    }
  }
  my @ideDevices    = grep { $_->{'interface'} eq 'ide' } @devices;
  my @scsiDevices   = grep { $_->{'interface'} eq 'scsi' } @devices;
  my @virtioDevices = grep { $_->{'interface'} eq 'virtio' } @devices;
  print("Block devices on $hostName:\n");
  printf('  %-'.($nameLength+2).'s%-11s%'.$indexLength.'s  %-7s%-7s%s'."\n", "Device", "Interface", "Index", "Media", "State", "File");
  foreach my $device ( @ideDevices, @scsiDevices, @virtioDevices) {
    printf('  %-'.($nameLength+2).'s%-11s%'.$indexLength.'s  %-7s%s%s%s    %s'."\n",
      $device->{'name'},
      $device->{'interface'},
      $device->{'idx'},
      $device->{'media'},
      ($device->{'removable'}?'r':'-'),
      ($device->{'locked'}?'l':'-'),
      ($device->{'inserted'}?'i':'-'),
      $device->{'file'});
  }
}

sub cmd_block_stats {
  my ($hostName, $format) = @_;
  fatal_if(!check_host_name($hostName), "Incorrect host name: $hostName\n", 1);
  fatal_if(!check_host_running($hostName), "Host does not exists.\n", 1);
  $format = '-b' unless( $format );
  given( $format) {
    when( "-b" )  { $format = 'b'; }
    when( "-k" )  { $format = 'k'; }
    when( "-m" )  { $format = 'm'; }
    when( "-g" )  { $format = 'g'; }
    when( "-t" )  { $format = 't'; }
    when( "-p" )  { $format = 'p'; }
    when( "-h" )  { $format = 'h'; }
    default       { fatal("Incorrect format\n", 1); }
  }
  my @devices = dev_block_list($hostName);
  @devices = sort { $a->{'idx'} cmp $b->{'idx'} } @devices;
  $nameLength = 6;
  $rdbLength = 10;
  $wrbLength = 11;
  $rdoLength = 10;
  $indexLength = $devices[-1]->{'idx'};
  $indexLength = length($indexLength);
  $indexLength = 5 if( $indexLength < 5 );
  foreach my $device ( @devices ) {
    if( length($device->{'name'}) > $nameLength ) {
      $nameLength = length($device->{'name'});
    }
    if( length(format_size($device->{'readBytes'}, $format)) > $rdbLength ) {
      $rdbLength = length(format_size($device->{'readBytes'}, $format));
    }
    if( length(format_size($device->{'writeBytes'}, $format)) > $wrbLength ) {
      $wrbLength = length(format_size($device->{'writeBytes'}, $format));
    }
    if( length($device->{'readOpers'}) > $rdoLength ) {
      $rdoLength = length($device->{'readOpers'});
    }
  }
  my @ideDevices    = grep { $_->{'interface'} eq 'ide' } @devices;
  my @scsiDevices   = grep { $_->{'interface'} eq 'scsi' } @devices;
  my @virtioDevices = grep { $_->{'interface'} eq 'virtio' } @devices;
  print("Block devices stats on $hostName:\n");
  printf('  %-'.($nameLength+2).'s%-11s%'.$indexLength.'s  %-7s%-7s%-'.($rdbLength+2).'s%-'.($wrbLength+2).'s%-'.($rdoLength+2).'s%s'."\n", "Device", "Interface", "Index", "Media", "State", "Read bytes", "Write bytes", "Read opers", "Write opers");
  foreach my $device ( @ideDevices, @scsiDevices, @virtioDevices) {
    printf('  %-'.($nameLength+2).'s%-11s%'.$indexLength.'s  %-7s%s%s%s    %-'.($rdbLength+2).'s%-'.($wrbLength+2).'s%-'.($rdoLength+2).'s%s'."\n",
      $device->{'name'},
      $device->{'interface'},
      $device->{'idx'},
      $device->{'media'},
      ($device->{'removable'}?'r':'-'),
      ($device->{'locked'}?'l':'-'),
      ($device->{'inserted'}?'i':'-'),
      format_size($device->{'readBytes'}, $format),
      format_size($device->{'writeBytes'}, $format),
      $device->{'readOpers'},
      $device->{'writeOpers'});
  }
}

sub cmd_version {
  my $version = VERSION;
  print "kvmsh v.$version\n";
  print "Copyright (C) 2009 by Mateusz Hromada\n";
}

sub cmd_help {
  my $version = VERSION;
  print "kvmsh v.$version\n";
  print "\n";
  print "Usage:\n";
  print "  kvmsh [action] [parameters]\n";
  print "  kvmsh [action] [host] [parameters]\n";
  print "\n";
  print "Actions:\n";
  print "  block-list host\n";
  print "  create host\n";
  print "  destroy host\n";
  print "  edit host\n";
  print "  help\n";
  print "  list [-auto|-host|-network]\n";
  print "  pause host\n";
  print "  reboot host\n";
  print "  show-config host\n";
  print "  shutdown host\n";
  print "  unpause host\n";
  print "  version\n";
}

my $user;
my $action;
my @args;

$user = getpwuid($<);
$user = $ENV{"SUDO_USER"} if( $user == "root" && $ENV{"SUDO_USER"} );

($action, @args) = @ARGV;

given( $action ) {
  when( "block-list" )  { cmd_block_list(@args);  }
  when( "block-stats" ) { cmd_block_stats(@args); }
  when( "create" )      { cmd_create(@args);      }
  when( "destroy" )     { cmd_destroy(@args);     }
  when( "edit" )        { cmd_edit(@args);        }
  when( "help" )        { cmd_help;               }
  when( "list" )        { cmd_list(@args);        }
  when( "pause" )       { cmd_pause(@args);       }
  when( "reboot" )      { cmd_reboot(@args);      }
  when( "show-config" ) { cmd_show_config(@args); }
  when( "shutdown" )    { cmd_shutdown(@args);    }
  when( "unpause" )     { cmd_unpause(@args);     }
  when( "version" )     { cmd_version;            }
  default               { cmd_help;               }
}
